import functools
import logging

import numpy

import OpenGL
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *


import pypatterns.filter as FilterModule

import pypatterns.relational as RelationalModule
import pomsets.resource as ResourceModule

import pomsets.definition as DefinitionModule
import pomsets.task as TaskModule
import pomsets.parameter as ParameterModule

import pomsets_app.gui.graph as GraphModule



class Canvas(ResourceModule.Struct):

    ATTRIBUTES = [
        'contextManager',
        'nodePolicy',
        'edgePolicy',
        'layoutPolicy',
        'visualPolicy',
        'portPolicy'
        ]

    def __init__(self):

        ResourceModule.Struct.__init__(self)

        self.scroll_position = [0.0, 0.0]

        # set up the zoom and pan matrices
        self.zoomMatrix = numpy.matrix('1 0 0 0; 0 1 0 0; 0 0 1 0; 0 0 0 1')
        self.panMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; 0 0 0 0')

        self.selection_rectangle = None

        self._selected = []
        self.resetDrawables()
        return


    def resetDrawables(self):
        self._dynamicEdge = None
        self.nodes = []
        self.edges = []
        self.selectables = set([])
        self.clickables = set([])
        self.layers = {}
        for layerName in ['edges', 'nodes']:
            layer = CanvasLayer(layerName)
            self.layers[layerName] = layer
            pass
        self.setSelection([])
        return


    def addDrawable(self, drawable):
        if isinstance(drawable, GraphModule.Node):
            self.nodes.append(drawable)
            self.layers['nodes'].addChild(drawable)

            ports = drawable.inputPorts.values() + drawable.outputPorts.values()
            for port in ports:
                self.determineSelectable(port)
                self.determineClickable(port)

        if isinstance(drawable, GraphModule.Edge):
            self.edges.append(drawable)
            self.layers['edges'].addChild(drawable)

        self.determineSelectable(drawable)
        self.determineClickable(drawable)
        return


    def determineSelectable(self, object):
        if object.isSelectable():
            self.selectables.add(object)
        return

    def determineClickable(self, object):
        if object.isClickable():
            self.clickables.add(object)

        return

    def removeDrawable(self, drawable):
        return self.removeChild(drawable)
    
    def removeChild(self, drawable):
        # remove from layers
        for layer in self.layers.values():
            layer.removeChild(drawable)

        if isinstance(drawable, GraphModule.Node):
            ports = drawable.inputPorts.values() + drawable.outputPorts.values()
            for port in ports:
                self.selectables.discard(port)
                self.clickables.discard(port)

        self.selectables.discard(drawable)
        self.clickables.discard(drawable)
        return

    def getSelectables(self):
        return self.selectables

    def getClickables(self):
        return self.clickables

    def computeLayout(self):

        if not len(self.nodes):
            # no need to compute layout if self has no nodes
            return

        (x_max, y_max) = self.layoutPolicy().layoutNodes(
            self.nodes, self.edges)
        self.scroll_bounds = (x_max, y_max)

        # because the node positions have been updated
        # the edge points need to be recomputed
        # map(lambda x: x.setupPrimitives(self.nodePolicy()), self.nodes)
        map(lambda x: x.setupPrimitives(self.edgePolicy()), self.edges)

        return


    def setScrollPosition(self, new_scroll_position):
        self.scroll_position = new_scroll_position
        self.validateScrollPosition()


    def computeDynamicEdge(self, initialPosition, finalPosition):

        edgePolicy = self.edgePolicy()

        initialCanvasMatrix = numpy.matrix(
            '1 0 0 0; 0 1 0 0; 0 0 1 0; %s %s 0 1' % 
            (initialPosition.x, initialPosition.y))
        finalCanvasMatrix = numpy.matrix(
            '1 0 0 0; 0 1 0 0; 0 0 1 0; %s %s 0 1' % 
            (finalPosition.x, finalPosition.y))

        initialWorldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            initialCanvasMatrix)
        finalWorldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(
            finalCanvasMatrix)

        point1 = GraphModule.Point(
            initialWorldMatrix[3,0], initialWorldMatrix[3,1])
        point2 = GraphModule.Point(
            finalWorldMatrix[3,0], finalWorldMatrix[3,1])

        line = GraphModule.Line()

        points = edgePolicy.createPath((point1.x, point1.y),
                                       (point2.x, point2.y))
        pointObjects = [GraphModule.Point(x,y,0) for x,y,z in points]
        line.points = pointObjects

        line.position = (0.0, 0.0, 0.01)
        line.color = [0.7, 0.7, 1.0, 0.3]
        
        edgeClass = self.contextManager().app().getResourceValue(
            'gui edge class', default=GraphModule.Edge)
        self._dynamicEdge = edgeClass()
        self._dynamicEdge.children = [line]
        self._dynamicEdge.setupColors(self.edgePolicy())
        # self.points = line.points

        pass


    def drawLayers(self):
        # Draw nodes
        for layerName, layer in self.layers.iteritems():
            for child in layer.getChildren():
                child.draw()
        return

    def drawDynamicEdge(self):
        if self._dynamicEdge is not None:
            self._dynamicEdge.draw()
        return

    def drawSelectionBox(self):
        # Draw selection rectangle
        if self.selection_rectangle is not None:
            glColor4f(0.8, 0.5, 0.2, 0.2)
            self.drawRect(self.selection_rectangle)
        return


    def draw(self):
        glPushMatrix()

        glTranslate(self.panMatrix[3,0], self.panMatrix[3,1], self.panMatrix[3,2])
        glScale(1.0/self.zoomMatrix[0,0], 1.0/self.zoomMatrix[1,1], 1.0/self.zoomMatrix[2,2])

        self.drawLayers()

        self.drawDynamicEdge()

        self.drawSelectionBox()

        glPopMatrix()
        return


    def getObjectsIntersectingRect(self, rect):

        # recalculate the rect so that its in the form 
        # where width and height are both positive
        baseX = rect.x
        baseY = rect.y
        diffX = baseX + rect.width
        diffY = baseY + rect.height

        minX = min(baseX, diffX)
        minY = min(baseY, diffY)
        maxX = max(baseX, diffX)
        maxY = max(baseY, diffY)
        # rect = (minX, minY, maxX-minX, maxY-minY)
        rect = GraphModule.Rect(minX, minY, maxX-minX, maxY-minY)

        objects = []
        for child in self.nodes + self.edges:
            if child.intersectsRect(rect):
                objects.append(child)
        return objects


    # END class Canvas
    pass




class GlCanvas(Canvas):
    
    ATTRIBUTES = Canvas.ATTRIBUTES + [
        'nodeTable',
    ]

    def __init__(self, parent, width, height, *args, **kwds):

        Canvas.__init__(self)

        self.width = width
        self.height = height

        self.initGL()
        self.initLighting()
        self.initViewport()

        self._frameIsInProcessOfShowing = False
        return

        return

    def resetDrawables(self):
        Canvas.resetDrawables(self)
        self.resetDataToGuiObjectMapping()

        # the check here is because
        # this is also being called in the constructor
        if self.nodeTable() is not None:
            self.nodeTable().clear()
        else:
            self.resetNodeTable()

        return


    def resetDataToGuiObjectMapping(self):
        self._dataToGuiObjectMapping = {}
        return

    def addDataToGuiObjectMapping(self, dataObject, guiObject):
        self._dataToGuiObjectMapping[dataObject] = guiObject
        return

    def getGuiObjectForDataObject(self, dataObject):
        return self._dataToGuiObjectMapping[dataObject]
 


    def getHeight(self):
        return self.height

    def getWidth(self):
        return self.width

    def resize(self, width, height):
        self.width = width
        self.height = height
        self.setViewport(width, height)
        return
    
    def initViewport(self):
        return self._initViewport(self.getWidth(), self.getHeight())

    def _initViewport(self, width, height):

        self.setViewport(width, height)
        self.panMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; 0 0 0 0')

        return

    def setViewport(self, width, height):
        # Reset The Current Viewport And Perspective Transformation
        glViewport(0, 0, width, height)		
        self._updateProjection(width, height)
        return

    def initLighting(self):
        return self._initLighting()

    def _initLighting(self):
        return


    def initGL(self):
        import platform
        if platform.system() in ['Linux']:
            glutInit([])
        return self._initGL()


    def _initGL(self):

        glClearDepth (1.0)
        glEnable (GL_DEPTH_TEST)
        # glClearColor (0.0, 0.0, 0.0, 0.0)
        glClearColor (0.6, 0.6, 0.6, 1.0)
        
        glShadeModel (GL_SMOOTH)
        glMatrixMode (GL_PROJECTION)

        # Enable blending
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Really Nice Perspective Calculations
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST) 

        # Enables Smooth Color Shading
        glShadeModel(GL_SMOOTH)
        # Enables Clearing Of The Depth Buffer
        glClearDepth(1.0)

        # Enables Depth Testing
        glEnable(GL_DEPTH_TEST)

        # The Type Of Depth Test To Do	
        glDepthFunc(GL_LEQUAL)

        glDisable(GL_LIGHTING)
        glColor3f(1.0, 1.0, 1.0)

        self.updateProjection()

        return


    def Render(self, dc):
        self.SetCurrent()

        self.displayFunc()

        self.SwapBuffers()
        return

    def displayFunc(self):
        return self._displayFunc()

    def _displayFunc(self):
        glDrawBuffer(GL_BACK)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glDisable(GL_LIGHTING)
        glColor3f(1.0, 1.0, 1.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        """main drawing stuff goes here"""
        self.draw()

        self.drawMetaInfo()
        return


    def drawMetaInfo(self):

        x = 30
        y = self.getHeight() - 30
        self.drawFPS(x, y)

        return


    def updateProjection(self):
        return self._updateProjection(self.getWidth(), self.getHeight())


    def _updateProjection(self, width, height):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(0.0, width, 0.0, height)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        return


    def drawFPS(self, x, y):
        return


    def getClickableObjectsAtCanvasCoordinates(self, x, y):
        canvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % (x, y))
        worldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(canvasMatrix)
        world_x = worldMatrix[3,0]
        world_y = worldMatrix[3,1]

        point = GraphModule.Point(world_x, world_y)

        for clickable in self.getClickables():
            if clickable.overlaps(point):
                yield clickable
        raise StopIteration


    def getSelectableObjectAtCanvasCoordinates(self, x, y):

        """
        this should instead be a composite filter
        # object is selectable
        # object contains point
        """

        canvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % (x, y))
        worldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(canvasMatrix)
        world_x = worldMatrix[3,0]
        world_y = worldMatrix[3,1]

        point = GraphModule.Point(world_x, world_y)

        for selectable in self.getSelectables():
            if selectable.overlaps(point):
                return selectable

        return None

    def getViewportCoordinatesFromCanvasCoordinates(self, canvasMatrix):
        rotationMatrix = numpy.matrix(
            '1 0 0 0; 0 -1 0 0; 0 0 -1 0; 0 0 0 1')
        translationMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; 0 %s 0 0' % self.getHeight())

        viewportMatrix = (canvasMatrix*rotationMatrix) +translationMatrix
        return viewportMatrix


    def getWorldCoordinatesFromViewportCoordinates(self, viewportMatrix):

        worldMatrix = self.zoomMatrix * (viewportMatrix - self.panMatrix)

        return worldMatrix


    def getWorldCoordinatesFromCanvasCoordinates(self, canvasMatrix):
        viewportMatrix = self.getViewportCoordinatesFromCanvasCoordinates(canvasMatrix)
        worldMatrix = self.getWorldCoordinatesFromViewportCoordinates(viewportMatrix)

        return worldMatrix


    """
    def setSelection(self, selected):

        oldSelection = self._selected
        self._selected = selected

        self.contextManager().selectionChanged(oldSelection, self._selected)

        return
    """

    """
    def addToSelection(self, selected):

        self._selected.extend(selected)

        self.contextManager().selectionChanged([], self._selected)
        pass
    """

    def getSelection(self):
        return self._selected


    def getResourceKeyForNode(self, dataNode):

        # TODO:
        # return the appropriate key
        # for the composite node,
        # e.g. nest, parameter sweep, loop, branch

        if not dataNode.isAtomic():
            generator = TaskModule.getTaskGeneratorForDefinition(dataNode)
            if isinstance(generator, TaskModule.NestTaskGenerator):
                return 'gui nest node class'
            #if isinstance(generator, TaskModule.ParameterSweepTaskGenerator):
            #    return 'gui parameter sweep node class'
            if isinstance(generator, TaskModule.LoopTaskGenerator):
                return 'gui loop node class'
            if isinstance(generator, TaskModule.BranchTaskGenerator):
                return 'gui branch node class'
            pass

        # default to generic node
        return 'gui node class'


    def addNode(self, dataNode, worldMatrix=None):

        resourceKey = self.getResourceKeyForNode(dataNode)
        nodeClass = self.contextManager().app().getResourceValue(
            resourceKey, default=GraphModule.Node)
        
        uiNode = nodeClass(dataNode)
        # need to do this so that the node
        # will have access to app contextual information
        uiNode.contextManager(self.contextManager())
        
        uiNode.setupPrimitives(self.nodePolicy())

        self.addDrawable(uiNode)

        if worldMatrix is None:
            worldMatrix = numpy.matrix('1 0 0 0; 0 1 0 0; 0 0 1 0; 0 0 0 1')

        uiNode.setPosition(worldMatrix[3,0], worldMatrix[3,1])

        dataObjectFunction = self.addDataToGuiObjectMapping

        dataObjectFunction(dataNode, uiNode)
        # ports = uiNode.inputPorts.values()+uiNode.outputPorts.values()
        for portDirection, ports in [('input', uiNode.inputPorts.values()),
                                  ('output',uiNode.outputPorts.values())]:
            for port in ports:
                portName = port.name

                # parameters have __eq__ defined
                # such that they are equivalent if their ids are equivalent
                # and since the ids are not unique (and are contextualized)
                # we need to include the context, ie the data node
                dataObjectFunction((dataNode, portName, portDirection), port)

            pass

        return uiNode


    def addEdge(self, dataEdge):
        # find the gui objects
        # for the outputNode, outputPort, inputPort, inputNode
        dataPath = dataEdge.path()

        dataObjectFunction = self.getGuiObjectForDataObject

        guiPath = map(
            dataObjectFunction,
            [
                dataPath[0], # source node
                (dataPath[0], dataPath[1], 'output'), # source port
                (dataPath[-1], dataPath[-2], 'input'), # target port
                dataPath[-1] # target node
            ]
        )

        edgeClass = self.contextManager().app().getResourceValue(
            'gui edge class', default=GraphModule.Edge)

        edge = edgeClass(dataPath, guiPath)
        edge.setupPrimitives(self.edgePolicy())
        self.addDrawable(edge)

        return

    def addDrawable(self, drawable):
        Canvas.addDrawable(self, drawable)

        # is it possible for the code that draws the nodes
        # to simply iterate over the rows of the table?
        if isinstance(drawable, GraphModule.Node):
            logging.debug("adding node %s to self.nodeTable()" % drawable.nodeData.name())
            row = self.nodeTable().addRow()
            row.setColumn('data', drawable.nodeData)
            row.setColumn('gui', drawable)
            pass

        pass


    def removeDrawable(self, drawable):
        if isinstance(drawable, GraphModule.Node):
            filter = RelationModule.ColumnValueFilter(
                'gui',
                FilterModule.IdentityFilter(drawable))
            self.nodeTable().removeRows(filter)
            pass

        Canvas.removeDrawable(self, drawable)
        return


    def computeLayout(self):
        Canvas.computeLayout(self)
        for uiNode in self.nodes:
            dataNode = uiNode.nodeData
            dataNode.guiOptions()['canvas position'] = uiNode.getPosition()
        return




    def resetNodeTable(self):
        table = RelationalModule.createTable(
            'node table',
            ['data', 'gui']
        )
        self.nodeTable(table)
        return

    def _getFilterForDataNode(self, node):
        filter = RelationalModule.ColumnValueFilter(
            'data',
            FilterModule.IdentityFilter(node)
        )
        return filter

    def hasGuiObjectForDataNode(self, node):
        filter = self._getFilterForDataNode(node)
        guiNodes = RelationalModule.Table.reduceRetrieve(
            self.nodeTable(),
            filter,
            ['gui']
        )
        if len(guiNodes) is 0:
            return False
        return True


    def getGuiObjectForDataNode(self, node):
        """
        should be able to use getGuiObjectForDataObject
        """

        filter = self._getFilterForDataNode(node)
        guiNodes = RelationalModule.Table.reduceRetrieve(
            self.nodeTable(),
            filter,
            ['gui']
        )
        if len(guiNodes) is 0:
            raise KeyError('no gui object found for node %s' % node.id())
        if len(guiNodes) is not 1:
            raise ValueError('expected only 1 gui object for node %s' % node.id())
        return guiNodes[0]


    def hasCanvasPosition(self, dataNode):
        guiOptions = dataNode.guiOptions()
        return 'canvas position' in guiOptions
    


    def updateExecutionStatus(self, task):
        # TODO:
        # this needs to handle the case that the task
        # is not pointing to the nest definition 
        # being displayed

        for uiNode in self.nodes:
            dataNode = uiNode.nodeData

            filter = task._getFilterForDefinition(dataNode)
            statusList = RelationalModule.Table.reduceRetrieve(
                task.tasksTable(),
                filter, ['status'], [])

            # update only if there's a status to update
            if not len(statusList):
                continue
            uiNode.updateExecuteStatus(statusList[0])
            pass

        self.updateGL()
        return


    def displayPomset(self, pomsetReferencePath=None):

        shouldComputeLayout = True
        shouldQueryUser = False
        
        # TODO:
        # determine if should query user
        self.resetDrawables()

        if pomsetReferencePath is None:
            pomsetContext, pomsetReferencePath, isEditable = \
                self.contextManager().currentDisplayedPomsetInfo()
            return
        pomsetReference = pomsetReferencePath[-1]
        pomset = pomsetReference.definitionToReference()

        # draw the nodes
        for dataNode in pomset.nodes():
            logging.debug('drawing node %s' % dataNode.name())
            uiNode = self.addNode(dataNode)
            if self.hasCanvasPosition(dataNode):
                shouldComputeLayout = False
                x, y = dataNode.guiOptions()['canvas position']
                uiNode.setPosition(x, y)
            else:
                shouldQueryUser = True
            pass

        # draw the execute status on the nodes
        if self.contextManager().hasTaskForPomsetPath(pomsetReferencePath):
            task = self.contextManager().getTaskForPomsetPath(pomsetReferencePath)
            self.updateExecutionStatus(task)
        
        # draw the edges between the nodes
        notNoneFilter = FilterModule.constructNotFilter()
        notNoneFilter.addFilter(
            RelationalModule.ColumnValueFilter(
                'edge',
                FilterModule.IdentityFilter(None)
            )
            )
        for values in pomset.parameterConnectionPathTable().retrieve(
            # FilterModule.TRUE_FILTER, 
            notNoneFilter,
            ['edge']):
            edge = values[0]
            self.addEdge(edge)
            pass

        # draw the pomset's parameters

        # draw the edges between the pomset's parameters
        # and node parameters

        if shouldQueryUser and not shouldComputeLayout:
            print "should query user for auto arrange"
            # TODO:
            # if user confirms auto-arrange
            # set shouldComputeLayout to True

        # TODO:
        # call computeLayout only if the loaded pomset
        # does not already contain information about node positioning
        # self.computeLayout()
        if shouldComputeLayout:
            self.computeLayout()


        return



    def breadCrumbs(self, value=None):
        if value is not None:
            self._breadCrumbs = value
        if not hasattr(self, '_breadCrumbs'):
            self._breadCrumbs = 'No active pomset to display'
        return self._breadCrumbs


    def setSelection(self, selected):

        for x in self.getSelection():
            x.isSelected(False)
            x.updateDisplay()

        self._selected = selected

        for x in selected:
            x.isSelected(True)
            x.updateDisplay()

        return

    def addToSelection(self, selected):

        self._selected.extend(selected)

        for x in selected:
            x.isSelected(True)
            x.updateDisplay()
        pass


    def drawLine(self, line):
        glBegin(GL_LINE_STRIP)
        for point in line.points:
            glVertex3f(point.x, point.y, point.z)
        glEnd()
        return


    def drawRect(self, rect, filled=True):
        glPushMatrix()
        glTranslatef(rect.x, rect.y, 0)
        if filled:
            glBegin(GL_QUADS)
        else:
            glBegin(GL_LINE_LOOP)
        glVertex2f(0, 0)
        glVertex2f(0, rect.height)
        glVertex2f(rect.width, rect.height)
        glVertex2f(rect.width, 0)
        glEnd()
        glPopMatrix()
        return

    # END class GlCanvas
    pass




class CanvasLayer(object):

    def __init__(self, name):
        self.name = name
        self._children = []
        return

    def addChild(self, child):
        self.getChildren().append(child)
        return

    def removeChild(self, child):
        while child in self.getChildren():
            self.getChildren().remove(child)
        return

    def getChildren(self):
        return self._children


    # END class CanvasLayer
    pass



"""
This is a setup.py script generated by py2applet

Usage:
    python setup.py py2app
"""

from setuptools import setup, find_packages
import py2app

APP_NAME = "pomsets"
BUNDLE_NAME = APP_NAME

# NOTE:
# remember to also update the Application package
VERSION="1.0.7"

APP = ['src/pomsets_app/pomsets_qt.py']
DATA_FILES = ['LICENSE']
RESOURCES = [
    'resources/config',
    'resources/fonts',
    'resources/images',
    'resources/licenses',
    'resources/library']
OPTIONS = {
    # specify icon file
    # iconfile='resources/myapp-icon.icns',
    # specify license
    # resources=['resources/License.txt'],
    'site_packages':True,
    'argv_emulation': False, 
    'includes': ['sip'],
    'resources':RESOURCES,
    'iconfile':'resources/images/logo.icns',
    'packages':[],
    'plist':dict(
        CFBundleName=BUNDLE_NAME,
        CFBundleShortVersionString=VERSION,
        CFBundleGetInfoString="%s %s" % (APP_NAME, VERSION),
        CFBundleExecutable=APP_NAME,
        CFBundleIdentifier="org.pomsets.app"
        )
    }

"""
    import py2app

    # Build the .app file
    setup(
        options=dict(
            py2app=dict(
                iconfile='resources/myapp-icon.icns',
                packages='wx',
                site_packages=True,
                resources=['resources/License.txt'],
                plist=dict(
                    CFBundleName               = "MyAppName",
                    CFBundleShortVersionString = "0.2.5",     # must be in X.X.X format
                    CFBundleGetInfoString      = "MyAppName 0.2.5",
                    CFBundleExecutable         = "MyAppName",
                    CFBundleIdentifier         = "com.example.myappname",
                ),
            ),
        ),
        app=[ 'myapp.py' ]
    )

"""



setup(
    name='pomsets-gui',
    version=VERSION,
    packages=find_packages(
        'src',exclude=["*.test", "*.test.*", "test.*", "test",
                       "*.utils", "*.utils.*", "utils.*", "utils",
                       "*.wx", "*.wx.*", "wx.*", "wx",
                       ]),
    package_dir={'':'src'},
    test_suite="test",
    tests_require=[
        "euca2ools>=1.1"
        ],
    install_requires = [
        'pypatterns>=0.1.1',
        'pomsets-core>=1.0.6',
        'PyOpenGL',
        ],
    app=APP,
    data_files=DATA_FILES,
    options={'py2app': OPTIONS},
    setup_requires=['py2app'],


    # metadata for upload to PyPI
    author = "michael j pan",
    author_email = "mikepan@gmail.com",
    description = "GUI for workflow management for the cloud",
    license = "GPL (personal use) or commercial",
    keywords = "workflow cloud hadoop parameter sweep",
    url = "http://pomsets.org",
    long_description = "pomsets-gui implements the GUI for workflow management in the cloud.",
    platforms = ["All"]

)



import logging
import os

from PyQt4.QtCore import *
from PyQt4 import QtGui

import pypatterns.filter as FilterModule

import pomsets.parameter as ParameterModule

import pomsets_app.utils as AppUtilsModule
import pomsets_app.gui.qt as QtModule

class Controller(QtModule.Controller, QtGui.QDialog):


    def __init__(self, *args, **kwds):
        QtGui.QDialog.__init__(self, *args, **kwds)
        QtModule.Controller.__init__(self)
        return

    def populate(self):

        widget = self.widget()
        contextManager = self.contextManager()

        automaton = contextManager.automaton()

        configuration = automaton.hadoopConfigurations().get('localhost', None)
        if configuration is None:
            configuration = {'hostname':'localhost'}

        textEditHadoopHome = widget.textEditHadoopHome
        textEditHadoopHome.setText(configuration.get('home', ''))

        checkBoxRelative = widget.checkBoxRelative
        isRelativeToHadoopHome = configuration.get(
            'streaming jar is relative to home', True)
        checkBoxRelative.setChecked(isRelativeToHadoopHome)

        textEditStreamingJar = widget.textEditStreamingJar

        defaultStreamingJar = ''
        if isRelativeToHadoopHome:
            defaultStreamingJar = os.path.join(
                'contrib', 'streaming',
                'hadoop-0.20.1-streaming.jar')
        else:
            defaultStreamingJar = os.path.join(
                defaultHadoopHome, 'contrib', 'streaming',
                'hadoop-0.20.1-streaming.jar')
        textEditStreamingJar.setText(
            configuration.get('streaming jar', defaultStreamingJar))

        return


    @pyqtSlot(QtGui.QAbstractButton)
    def on_buttonBox_clicked(self, button):
        widget = self.widget()
        buttonRole = widget.buttonBox.buttonRole(button)
        if buttonRole == QtGui.QDialogButtonBox.ApplyRole:
            self.saveValues()
            pass
        if buttonRole == QtGui.QDialogButtonBox.AcceptRole:
            self.saveValues()
            self.accept()
            pass
        if buttonRole == QtGui.QDialogButtonBox.RejectRole:
            self.reject()
            pass
        return


    @pyqtSlot()
    def on_buttonBrowse_clicked(self):
        widget = self.widget()
        textEditStreamingJar = widget.textEditStreamingJar

        import user
        initialDir = user.home
        streamingJar = str(textEditStreamingJar.text())
        if len(streamingJar):
            index = streamingJar.rfind(os.path.sep)
            if not index == -1:
                initialDir = streamingJar[:index]
            pass

        fileName = QtGui.QFileDialog.getOpenFileName(
            self,
            QString("Hadoop streaming jar file"), 
            QString(initialDir),
            QString("Jar file (*.jar)"))
        if fileName and len(fileName):
            textEditStreamingJar.setText(fileName)
        return


    def saveValues(self):

        contextManager = self.contextManager()
        automaton = contextManager.automaton()
        widget = self.widget()

        configuration = automaton.hadoopConfigurations().get('localhost', None)
        shouldCreateNewConfig = False
        if configuration is None:
            configuration = {'hostname':'localhost'}
            shouldCreateNewConfig = True

        textEditHadoopHome = widget.textEditHadoopHome
        configuration['home'] = str(textEditHadoopHome.text())


        checkBoxRelative = widget.checkBoxRelative
        isRelativeToHadoopHome = checkBoxRelative.isChecked()
        configuration['streaming jar is relative to home'] = isRelativeToHadoopHome

        textEditStreamingJar = widget.textEditStreamingJar
        configuration['streaming jar'] = str(textEditStreamingJar.text())

        if shouldCreateNewConfig:
            logging.debug("need to implement when no configuration found")
            print "need to implement when no configuration found"
            automaton.addHadoopConfiguration(configuration)

        # save out to file
        configData = AppUtilsModule.createNewConfigDataObject()
        automaton.addConfigToSave(configData)
        AppUtilsModule.saveConfig(configData)

        return


    # END class Controller
    pass



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/qt/edit parameter dialog.ui'
#
# Created: Thu Mar  4 16:12:54 2010
#      by: PyQt4 UI code generator 4.7
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(402, 439)
        self.buttonBox = QtGui.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(100, 390, 281, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Apply|QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.labelName = QtGui.QLabel(Dialog)
        self.labelName.setGeometry(QtCore.QRect(20, 30, 62, 16))
        self.labelName.setObjectName("labelName")
        self.labelDescription = QtGui.QLabel(Dialog)
        self.labelDescription.setGeometry(QtCore.QRect(20, 70, 81, 16))
        self.labelDescription.setObjectName("labelDescription")
        self.labelFlag = QtGui.QLabel(Dialog)
        self.labelFlag.setGeometry(QtCore.QRect(20, 190, 62, 16))
        self.labelFlag.setObjectName("labelFlag")
        self.labelDirection = QtGui.QLabel(Dialog)
        self.labelDirection.setGeometry(QtCore.QRect(20, 110, 62, 16))
        self.labelDirection.setObjectName("labelDirection")
        self.comboBoxDirection = QtGui.QComboBox(Dialog)
        self.comboBoxDirection.setGeometry(QtCore.QRect(120, 105, 261, 31))
        self.comboBoxDirection.setObjectName("comboBoxDirection")
        self.checkBoxCommandline = QtGui.QCheckBox(Dialog)
        self.checkBoxCommandline.setGeometry(QtCore.QRect(20, 150, 131, 20))
        self.checkBoxCommandline.setObjectName("checkBoxCommandline")
        self.checkBoxOptional = QtGui.QCheckBox(Dialog)
        self.checkBoxOptional.setGeometry(QtCore.QRect(20, 230, 131, 20))
        self.checkBoxOptional.setObjectName("checkBoxOptional")
        self.checkBoxFile = QtGui.QCheckBox(Dialog)
        self.checkBoxFile.setGeometry(QtCore.QRect(20, 270, 131, 20))
        self.checkBoxFile.setObjectName("checkBoxFile")
        self.checkBoxList = QtGui.QCheckBox(Dialog)
        self.checkBoxList.setGeometry(QtCore.QRect(20, 310, 131, 20))
        self.checkBoxList.setObjectName("checkBoxList")
        self.checkBoxDistributeFlag = QtGui.QCheckBox(Dialog)
        self.checkBoxDistributeFlag.setGeometry(QtCore.QRect(20, 350, 131, 20))
        self.checkBoxDistributeFlag.setObjectName("checkBoxDistributeFlag")
        self.textEditName = QtGui.QLineEdit(Dialog)
        self.textEditName.setGeometry(QtCore.QRect(120, 21, 261, 31))
        self.textEditName.setObjectName("textEditName")
        self.textEditDescription = QtGui.QLineEdit(Dialog)
        self.textEditDescription.setGeometry(QtCore.QRect(120, 60, 261, 31))
        self.textEditDescription.setObjectName("textEditDescription")
        self.textEditFlag = QtGui.QLineEdit(Dialog)
        self.textEditFlag.setGeometry(QtCore.QRect(120, 180, 261, 31))
        self.textEditFlag.setObjectName("textEditFlag")

        self.retranslateUi(Dialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), Dialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtGui.QApplication.translate("Dialog", "Edit parameter", None, QtGui.QApplication.UnicodeUTF8))
        self.labelName.setText(QtGui.QApplication.translate("Dialog", "Name", None, QtGui.QApplication.UnicodeUTF8))
        self.labelDescription.setText(QtGui.QApplication.translate("Dialog", "Description", None, QtGui.QApplication.UnicodeUTF8))
        self.labelFlag.setText(QtGui.QApplication.translate("Dialog", "Flag", None, QtGui.QApplication.UnicodeUTF8))
        self.labelDirection.setText(QtGui.QApplication.translate("Dialog", "Direction", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxCommandline.setText(QtGui.QApplication.translate("Dialog", "Commandline", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxOptional.setText(QtGui.QApplication.translate("Dialog", "Optional", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxFile.setText(QtGui.QApplication.translate("Dialog", "File", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxList.setText(QtGui.QApplication.translate("Dialog", "List", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxDistributeFlag.setText(QtGui.QApplication.translate("Dialog", "Distribute flag", None, QtGui.QApplication.UnicodeUTF8))


if __name__ == "__main__":
    import sys
    app = QtGui.QApplication(sys.argv)
    Dialog = QtGui.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())




import logging
import os
import platform
import shutil

from PyQt4 import QtGui
from PyQt4.QtCore import *

import pypatterns.command as CommandPatternModule

import cloudpool as CloudModule
import cloudpool.shell as ShellModule


import pomsets.library as LibraryModule
import pomsets.resource as ResourceModule


import pomsets_app.utils as AppUtilsModule
import pomsets_app.controller.automaton as AutomatonModule
import pomsets_app.controller.context as ContextModule

import pomsets_app.gui.graph as GraphModule

import pomsets_app.gui.qt.frame as FrameModule
import pomsets_app.gui.qt.menu as MenuModule
import pomsets_app.gui.policy as PolicyModule




def initializeAppSettings(config):

    if not 'application settings' in config:
        config['application settings'] = {}

    applicationSettings = config['application settings']

    # we need to make this distinction between how nodes are created
    # because linux does not seem to handle end drag events
    # and because macs typically don't have right mouse buttons
    if platform.system() in ['Darwin']:
        applicationSettings['create node via dnd'] = True
        applicationSettings['create node via canvas contextual menu'] = False
        pass
    elif platform.system() in ['Linux']:
        applicationSettings['create node via dnd'] = False
        applicationSettings['create node via canvas contextual menu'] = True
        pass
    return



def initializeAppConfig(resourcePath):

    configPath = AppUtilsModule.getDefaultConfigPath()

    # first, copy the config file
    shutil.copyfile(os.path.join(resourcePath, 'config', 'config'),
                    configPath)

    # now add the platform specific stuff
    config = AppUtilsModule.loadConfig()
    initializeAppSettings(config)

    AppUtilsModule.saveConfig(config, configPath=configPath)
    return


def getDefaultResourcePath():

    # TODO:
    # need to figure out how to determine the application's
    # resource path on different platforms

    endIndex = __file__.rfind('src')
    if endIndex == -1:
        # this is production (not development mode)
        endIndex = __file__.rfind('pomsets-gui')

    appRoot = __file__[:endIndex]
    if appRoot.endswith(os.path.sep):
        appRoot = appRoot[:-1]

    if platform.system() in ['Darwin']:

        hasDeterminedResourcePath = False

        applicationFilePath = str(QtGui.QApplication.applicationFilePath())

        pathToTest = 'pomsets-gui.app/Contents/MacOS' 
        if pathToTest in applicationFilePath:
            # we're using the application bundle
            resourcePath = applicationFilePath[:applicationFilePath.index(pathToTest)+len(pathToTest)-len('MacOS')] + 'Resources'
            hasDeterminedResourcePath = True
            pass
            
        pathToTest = 'pomsets.app/Contents/MacOS'
        if not hasDeterminedResourcePath and pathToTest in applicationFilePath:
            # we're using the application bundle
            resourcePath = applicationFilePath[:applicationFilePath.index(pathToTest)+len(pathToTest)-len('MacOS')] + 'Resources'
            hasDeterminedResourcePath = True
            pass

        if not hasDeterminedResourcePath:
            # we're not in the bundle, so a dev env,
            # use the dev default
            resourcePath = os.path.join(os.getcwd(), appRoot, 'resources')

        pass
    else:
        resourcePath = os.path.join(os.getcwd(), appRoot, 'resources')


    return resourcePath




class Application(QtGui.QApplication, ResourceModule.Struct):

    ATTRIBUTES = ['contextManager']

    def __init__(self):
        QtGui.QApplication.__init__(self, [])
        ResourceModule.Struct.__init__(self)

        self.setApplicationName('pomsets')
        self.setApplicationVersion('1.0.7')

        self._resourceValues = {}

        return


    def setResourceValue(self, key, value):
        self._resourceValues[key] = value
        return

    def getResourceValue(self, key, default=None):
        if not key in self._resourceValues:
            return default
        return self._resourceValues[key]

    def setDefaultResourcePath(self):
        self.setResourcePath(getDefaultResourcePath())
        return

    def setResourcePath(self, resourcePath):
        if not os.path.exists(resourcePath):
            raise IOError('resource path %s does not exist' % resourcePath)
        self._resourcePath = resourcePath
        return


    def runProgram(self):
        resourcePath = getDefaultResourcePath()
        self.setResourcePath(resourcePath)

        self.createApplicationContext()
        self.createApplicationFrame()

        if self.shouldQueryUserForLicenseAgreement():
            self.showLicenseFrame()
            self.exec_()

        if self.userHasAgreedToLicense():
            self.initializeForMainProgram()
            self.showMainFrame()
            self.exec_()

            # teardown, now that control of the event loop
            # has been returned to us
            automaton = self.contextManager().automaton()
            automaton.shouldProcessEndedTasks(False)
            return

        return



    def showSplashScreen(self):

        resourcePath = getDefaultResourcePath()
        # splashImage = os.path.join(resourcePath, 'images', 'logo.png')

        pixmap = QtGui.QPixmap(splashImage)
        splash = QtGui.QSplashScreen(pixmap)
        splash.show()

        # if mac, will need to call raise
        splash.raise_()
        return



    def showLicenseFrame(self):
        frame = self.getLicenseFrame(self.contextManager().mainWindow())
        frame.show()
        return

    def showMainFrame(self):
        self.contextManager().mainWindow().show()
        return

    def shouldQueryUserForLicenseAgreement(self):
        if self.userHasAgreedToLicense():
            return False
        return True

    def getLicenseFile(self):
        configDir = AppUtilsModule.getDefaultConfigDir()
        licenseFile = os.path.join(configDir, '.license')
        return licenseFile

    def userHasAgreedToLicense(self):
        licenseFile = self.getLicenseFile()
        return os.path.exists(licenseFile)


    def getLicenseFrame(self, parentWindow):
        contextManager = self.contextManager()

        # import the modules
        import pomsets_app.gui.qt.license_agreement.widget as WidgetModule
        import pomsets_app.gui.qt.license_agreement.controller as ControllerModule

        # initialize the widget
        controller = ControllerModule.Controller(parentWindow)
        controller.contextManager(contextManager)

        ui = WidgetModule.Ui_Dialog()
        ui.setupUi(controller)

        controller.widget(ui)

        # populate the dialog
        controller.populate()

        return controller



    def initializeForMainProgram(self):
        
        self.contextManager().initializeLibraries()
        self.initializeResources()
        self.initializeAutomaton()
        self.initializeFrame()
        self.createInitialPomsetContext()
        return



    def initializeResources(self):


        # need to ensure that the config file exists
        configPath = AppUtilsModule.getDefaultConfigPath()
        if not os.path.exists(configPath):
            configDir = os.path.dirname(configPath)
            if not os.path.exists(configDir):
                os.makedirs(configDir)
            resourcePath = getDefaultResourcePath()
            initializeAppConfig(resourcePath)

        config = AppUtilsModule.loadConfig()

        if not 'application settings' in config:
            initializeAppSettings(config)

        applicationSettings = config['application settings']

        if applicationSettings.get('create node via canvas contextual menu', False):
            self.setResourceValue('canvas contextual menu class',
                                 MenuModule.CanvasContextualMenuWithCreateNodeEnabled)
        else:
            self.setResourceValue('canvas contextual menu class',
                                 MenuModule.CanvasContextualMenu)

        self.setResourceValue('canvas contextual menu class',
                             MenuModule.CanvasContextualMenuWithCreateNodeEnabled)


        self.setResourceValue('gui node class',
                             GraphModule.Node)
        self.setResourceValue('gui nest node class',
                             GraphModule.NestNode)
        self.setResourceValue('gui loop node class',
                             GraphModule.LoopNode)
        self.setResourceValue('gui branch node class',
                             GraphModule.BranchNode)
        self.setResourceValue('gui edge class',
                             GraphModule.Edge)
        self.setResourceValue('gui port class',
                             GraphModule.Port)


        return
        


    def initializePolicies(self):

        self.setResourceValue('node policy class',
                             PolicyModule.NodePolicy)
        self.setResourceValue('visual policy class',
                             PolicyModule.VisualPolicy)

        config = AppUtilsModule.loadConfig()

        layoutClass = PolicyModule.BasicLayoutPolicy
        """
        try:
            import gv

            # TODO:
            # need to spcify the Python API version
            print "imported gv"
            layoutClass = PolicyModule.LayoutPolicyModule.GraphVizLayoutPolicy
        except ImportError:
            graphvizConfigurations = config.get('graphviz configurations', [])
            if len(graphvizConfigurations):
                graphvizConfiguration = graphvizConfigurations[0]
                dotCommandPath = graphvizConfiguration.get('dot command path')
                if dotCommandPath and os.path.exists(dotCommandPath):
                    print "found graphviz configuration"
                    layoutClass = PolicyModule.LayoutPolicyModule.GraphVizLayoutPolicy
                pass
            pass
        """
        graphvizConfigurations = config.get('graphviz configurations', [])
        if len(graphvizConfigurations):
            graphvizConfiguration = graphvizConfigurations[0]
            dotCommandPath = graphvizConfiguration.get('dot command path')
            if dotCommandPath and os.path.exists(dotCommandPath):
                layoutClass = PolicyModule.LayoutPolicyModule.GraphVizLayoutPolicy
                try:
                    import gv
                    # import succeeded
                    # so we can use gv_python to generate 
                    # the dot file with the non-positioned nodes
                    layoutClass = PolicyModule.GraphvizLayoutPolicy
                except ImportError:
                    # gv_python not installed
                    # use default one where we generate the dot file
                    pass
            pass

        self.setResourceValue('layout policy class', layoutClass)

        contextManager = self.contextManager()

        nodePolicy = self.initializeNodePolicy(contextManager)
        edgePolicy = self.initializeEdgePolicy(contextManager)
        visualPolicy = self.initializeVisualPolicy(contextManager)
        layoutPolicy = self.initializeLayoutPolicy(contextManager)
        portPolicy = self.initializePortPolicy(contextManager)
        nodePolicy.portPolicy(portPolicy)
        return


    def initializeFonts(self):

        # TODO:
        # this was copied from, and should use
        # the visual policy function instead
        fontsDir = os.sep.join([self.contextManager().resourcePath(),
                                'fonts'])
        fontPath = os.sep.join([fontsDir, 'FreeUniversal-Regular.ttf'])
        fontId = QtGui.QFontDatabase.addApplicationFont(fontPath)
        if fontId is not -1:
            self._defaultFontId = fontId
            fontFamilies = QtGui.QFontDatabase.applicationFontFamilies(fontId)
            # returns 'FreeUniversal'
        self._fontDb = QtGui.QFontDatabase()
        fontStyles = self._fontDb.styles('FreeUniversal')
        return


    def createApplicationFrame(self):
        frame = FrameModule.Frame()
        frame.app(self)
        self.contextManager().mainWindow(frame)
        return frame

    def initializeFrame(self):
        frame = self.contextManager().mainWindow()
        frame.initializeWidgets()
        frame.setWidgetDataSources()
        frame.populateWidgets()

        self.initializeCanvas(frame)
        return


    def initializeCanvas(self, frame):

        canvas = frame.canvas
        contextManager = self.contextManager()

        # contextManager.initializeCanvas(canvas=canvas)
        # contextManager.initializeEventHandlers()
        contextManager.canvas = canvas

        self.initializePolicies()
        self.initializeFonts()

        return


    def createInitialPomsetContext(self):
        contextManager = self.contextManager()

        newPomsetContext = contextManager.createNewPomset(name="new pomset")
        newPomsetContext.isModified(False)
        contextManager.addActivePomsetContext(newPomsetContext)

        frame = contextManager.mainWindow()
        frame.displayPomset(newPomsetContext)
        frame.emit(SIGNAL("OnPomsetLoaded(PyQt_PyObject)"), newPomsetContext)

        return newPomsetContext




    def createApplicationContext(self):
        contextManager = ContextModule.QtContextManager()        

        self.contextManager(contextManager)
        contextManager.app(self)

        # set the location of the resources for this application
        if not hasattr(self, '_resourcePath'):
            raise AttributeError('need to have set the resource path')
        self.contextManager().resourcePath(self._resourcePath)


        return


    def initializeAutomaton(self):

        automaton = AutomatonModule.Automaton()

        contextManager = self.contextManager()

        contextManager.initializeAutomaton(automaton)

        config = AppUtilsModule.loadConfig()
        automaton.loadConfig(config)


        # determine if graphviz is specified
        dotCommandPath = ''
        graphvizConfigurations = automaton.otherConfigurations().get(
            'graphviz configurations')
        if graphvizConfigurations and len(graphvizConfigurations):
            graphvizConfiguration = graphvizConfigurations[0]
            dotCommandPath = graphvizConfiguration.get('dot command path', '')
            pass
        contextManager.commandPath('dot', dotCommandPath)
        logging.debug('using %s as the dot command path' % dotCommandPath)


        # initialize the threadpool
        # this is for testing vm execution
        # should start with 0

        # this is for local execution
        threadpool = CloudModule.Pool(0)
        automaton.setThreadPool(AutomatonModule.ID_EXECUTE_LOCAL, threadpool)
        worker = threadpool.assignWorker()
        shell = ShellModule.LocalShell()
        worker.executeEnvironment(shell)

        # this is for remote execution
        credentials = automaton.remoteExecuteCredentials()
        if len(credentials):
            threadpool = CloudModule.Pool(0)
            automaton.setThreadPool(AutomatonModule.ID_EXECUTE_REMOTE, threadpool)

            credential = credentials[0]
            # TODO:
            # should wait until the user
            hostname = credential['hostname']
            userLogin = credential['user']
            keyfile = credential['keyfile']
            worker = threadpool.assignWorker()
            shell = ShellModule.SecureShell()
            shell.hostname(hostname)
            shell.user(userLogin)
            shell.keyfile(keyfile)
            worker.executeEnvironment(shell)

        """
        credentialEntries = [
            x for x in automaton.getCloudControllerCredentials(
                columns=AutomatonModule.Automaton.COLUMNS_CLOUDCONTROLLERCREDENTIALS)
        ]
        if len(credentialEntries):
            for credentialEntry in credentialEntries:
                credentials = credentialEntry[2]
                api = credentialEntry[1]
                name = credentialEntry[0]
                if api == 'euca2ools':
                    # this is for cloud execution
                    threadpool = CloudModule.Pool(0)
                    automaton.setThreadPool(AutomatonModule.ID_EXECUTE_EUCA2OOLS, threadpool)
                    # threadpool.credentials(credentials)
                else:
                    logging.warn('not implemented for %s API' % api)


                # serviceName
                # serviceAPI
                # values
                pass
            pass
        """
        threadpool = CloudModule.Pool(0)
        automaton.setThreadPool(AutomatonModule.ID_EXECUTE_EUCA2OOLS, threadpool)


        automaton.commandManager(CommandPatternModule.CommandManager())

        # startup the thread
        automaton.startProcessEndedTasksThread()

        library = contextManager.persistentLibrary()

        """
        # TODO:
        # should consolidate with the code in pomsets-core

        definition = library.getBootstrapLoader()
        executeEnvironmentMap = {
            'library bootstrap loader':LibraryModule.LibraryLoader(library)
            }

        commandBuilder = LibraryModule.CommandBuilder()
        commandBuilderMap = {
            'library bootstrap loader':commandBuilder,
            # 'python eval':commandBuilder
        }

        requestKwds = contextManager.generateRequestKwds(
            executeEnvironmentMap=executeEnvironmentMap,
            executeEnvironmentId = AutomatonModule.ID_EXECUTE_LOCAL,
            commandBuilderMap=commandBuilderMap)

        compositeTask = contextManager.executePomset(
            pomset=definition,
            requestKwds=requestKwds)
        """
        request = automaton.runBootstrapLoader(library, isCritical=False)
        task = request.kwds['task']

        # handle the case where there are errors on loading
        childTasks = task.getChildTasks()
        failedChildTasks = [x for x in childTasks if 
                            x.workRequest().exception]
        if len(failedChildTasks):
            #raise NotImplementedError(
            #    'should query user to remove from library')
            logging.error('should query user to remove from library')

        return automaton



    def initializeEdgePolicy(self, contextManager):
        canvas = contextManager.canvas
        
        # create and set the edge policy
        key = 'edge policy'
        classObject = self.getResourceValue(
            "%s class" % key, default=PolicyModule.EdgePolicy)
        edgePolicy = classObject()
        edgePolicy.contextManager(contextManager)
        canvas.edgePolicy(edgePolicy)
        

        visualPolicy = canvas.visualPolicy()
        edgePolicy.visualPolicy(visualPolicy)



        return edgePolicy
    

    def initializeLayoutPolicy(self, contextManager):
        canvas = contextManager.canvas
        
        # create and set the layout policy
        key = 'layout policy'
        classObject = self.getResourceValue(
            "%s class" % key, default=PolicyModule.GraphvizLayoutPolicy)
        layoutPolicy = classObject()
        layoutPolicy.contextManager(contextManager)
        canvas.layoutPolicy(layoutPolicy)

        return layoutPolicy
    

    def initializeNodePolicy(self, contextManager):
        canvas = contextManager.canvas

        # create the node policy for this application
        key = 'node policy'
        classObject = self.getResourceValue(
            "%s class" % key, default=PolicyModule.NodePolicy)
        nodePolicy = classObject()
        nodePolicy.contextManager(contextManager)

        # set the node policy
        canvas.nodePolicy(nodePolicy)
        return nodePolicy
    
    
    
    def initializeVisualPolicy(self, contextManager):

        # create the visual policy
        key = 'visual policy'
        classObject = self.getResourceValue(
            "%s class" % key, default=PolicyModule.VisualPolicy)
        visualPolicy = classObject()
        visualPolicy.contextManager(contextManager)
            
        # seupt the default font
        pathForDefaultFont = visualPolicy.getPathForDefaultFont()
        
        canvas = contextManager.canvas

        # set the visual policy
        canvas.visualPolicy(visualPolicy)
        

        # set the visual policy on the node policy
        nodePolicy = canvas.nodePolicy()
        nodePolicy.visualPolicy(visualPolicy)

        # set the visual policy on the edge policy
        edgePolicy = canvas.edgePolicy()
        edgePolicy.visualPolicy(visualPolicy)



        # set the colors for the node policy
        # we can instead set it on the visual policy first
        for key, value in [
            (PolicyModule.NodePolicy.KEY_BORDER_SELECTION_TRUE,  [1.0, 1.0, 1.0, 1.0]),
            (PolicyModule.NodePolicy.KEY_BORDER_SELECTION_FALSE,  [0.6, 0.6, 0.6, 1.0]),
            (PolicyModule.NodePolicy.KEY_BACKGROUND_SELECTION_TRUE,  [0.4, 0.4, 0.6, 1.0]),
            (PolicyModule.NodePolicy.KEY_BACKGROUND_SELECTION_FALSE,  [0.3, 0.3, 0.5, 0.8]),
            (PolicyModule.NodePolicy.KEY_TEXT_SELECTION_TRUE, [1.0, 1.0, 1.0, 1.0]),
            (PolicyModule.NodePolicy.KEY_TEXT_SELECTION_FALSE, [0.8, 0.8, 0.8, 1.0])]:

            nodePolicy.color(key, value)
            pass


        # set the colors for the edge policy
        # we can instead set it on the visual policy first
        for key, value in [
            (PolicyModule.EdgePolicy.KEY_DYNAMIC,  [0.85, 0.85, 0.85, 1.0]),
            (PolicyModule.EdgePolicy.KEY_SELECTION_TRUE, [0.85, 0.85, 0.0, 1.0]),
            (PolicyModule.EdgePolicy.KEY_SELECTION_FALSE, [0.7, 0.7, 0.0, 1.0])]:

            visualPolicy.color(key, value)
            pass

        return visualPolicy
    

    def initializePortPolicy(self, contextManager):

        key = 'port policy'
        classObject = self.getResourceValue(
            "%s class" % key, default=PolicyModule.PortPolicy)
        portPolicy = classObject(contextManager)

        canvas = contextManager.canvas
        canvas.portPolicy(portPolicy)
        return portPolicy


    # END class Application
    pass



import wx
import pomsets_app.gui.widget as WidgetModule

class VerifyExitDialog(WidgetModule.Dialog):

    XRC_NAME = 'VerifyExitDialog'

    WIDGET_BINDINGS = WidgetModule.Dialog.WIDGET_BINDINGS + [
        ('wxID_OK', 'OnButtonOk', wx.EVT_BUTTON),
    ]

    def __init__(self, xmlResource, parent=None):

        WidgetModule.Dialog.__init__(
            self, 
            xmlResource.LoadDialog(parent, VerifyExitDialog.XRC_NAME))

        self.SetSize(wx.Size(300, 300))
        #self.SetAutoLayout(True)


        messageTextCtrl = self.getWidget('messageTextCtrl')
        messageTextCtrl.SetMinSize(wx.Size(275,220))
        # messageTextCtrl.SetSize(wx.Size(250,280))
        messageTextCtrl.SetValue(
"""Thare are unsaved pomsets.
Do you wish to continue exiting the application?
Press OK to exit without saving, Cancel to cancel exit"""
)

        return

    def unsavedPomsetContexts(self, value=None):
        if value is not None:
            self._unsavedPomsetContexts = value
        if not hasattr(self, '_unsavedPomsetContexts'):
            self._unsavedPomsetContexts = []
        return self._unsavedPomsetContexts


    # END class VerifyExitDialog
    pass




import logging
import os

import pypatterns.filter as FilterModule
import pypatterns.relational as RelationalModule


import pomsets.definition as DefinitionModule
import pomsets.task as TaskModule
import pomsets.parameter as ParameterModule
import pomsets.resource as ResourceModule


from PyQt4.QtCore import *


import pomsets_app.gui.policy as PolicyModule

IMAGE_SELECTED_TRUE = 'image selected'
IMAGE_SELECTED_FALSE = 'image not selected'
IMAGE_PARAMETER_SWEEP = 'image parameter sweep'


class LocalGraphObject(ResourceModule.Struct):

    ATTRIBUTES = [
        'visible', 'isSelected', 'isSelectable', 'isClickable',
        'contextManager'
        ]

    # This is a hack,
    # until we move the common graph object code 
    def __init__(self):
        ResourceModule.Struct.__init__(self)
        self.isSelected(False)
        self.visible(True)
        return

    def overlaps(self, point):
        return False

    pass


class Point(object):
    def __init__(self, x, y, z=0):
        self.x = x
        self.y = y
        self.z = z
        return

class Rect(LocalGraphObject):

    ATTRIBUTES = LocalGraphObject.ATTRIBUTES + [
        'filled'
        ]

    def __init__(self, x, y, width, height):
        LocalGraphObject.__init__(self)
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        return

    def containsPoint(self, point):
        # TODO:
        # need to handle negative widths and heights

        return self.x < point.x and \
            self.x + self.width > point.x and \
            self.y < point.y and \
            self.y + self.height > point.y


    def intersectsLine(self, line):
        points = [Point(self.x, self.y),
                  Point(self.x+self.width, self.y),
                  Point(self.x+self.width, self.y+self.height),
                  Point(self.x, self.y+self.height)]
        for points in zip(points[:-1],points[1:]):
            rectEdge = Line()
            rectEdge.points = points
            if line.intersectsLine(rectEdge):
                return True
        return False

    # END class Rect
    pass



class Line(LocalGraphObject):

    ATTRIBUTES = LocalGraphObject.ATTRIBUTES + []

    @staticmethod
    def ccw(A,B,C):
        """
        from http://www.bryceboe.com/2006/10/23/line-segment-intersection-algorithm/
        """
	return (C.y-A.y)*(B.x-A.x) > (B.y-A.y)*(C.x-A.x)

    def __init__(self):
        LocalGraphObject.__init__(self)
        return
    
    def intersectsLine(self, other):
        A = self.points[0]
        B = self.points[1]
        C = other.points[0]
        D = other.points[1]
        return Line.ccw(A,C,D) != Line.ccw(B,C,D) and \
            Line.ccw(A,B,C) != Line.ccw(A,B,D)
        
    # END class Line
    pass


class Triangle(object):
    # END class Triangle
    pass


class Circle(object):

    def __init__(self, x, y, radius):
        self.x = x
        self.y = y
        self.radius = radius
        return

    # END class Circle
    pass


class Text(LocalGraphObject):
    ALIGN_LEFT = 'left'
    ALIGN_RIGHT = 'right'
    ALIGN_CENTER = 'center'

    ALIGN_TOP = 'top'
    ALIGN_MIDDLE = 'middle'
    ALIGN_BOTTOM = 'bottom'

    def __init__(self):
        LocalGraphObject.__init__(self)
        return

    """
    def getDrawingOffset(self, textBoundingRect, zoomFactor):

        # TODO:
        # use QFontMetrics.boundingRect


        # fontSize = self.font.getSize(self.text)

        selfWidth = self.size[0]
        selfHeight = self.size[1]

        xx = 0
        if self.horizontal_align is Text.ALIGN_CENTER:		# CENTER ALIGN
            print "in center halign"
            xx = (selfWidth-textBoundingRect.width())/2
        elif self.horizontal_align is Text.ALIGN_RIGHT:	# RIGHT ALIGN
            print "in right halign"
            xx = (selfWidth-textBoundingRect.width())
        else:
            print "in left halign"
    
        #yy = zoomFactor * textBoundingRect.height()/4.0
        # yy = 0
        yy = textBoundingRect.height() / 4.0
        if self.vertical_align is Text.ALIGN_MIDDLE:		# MIDDLE ALIGN
            print "in middle valign"
            yy = yy+ (selfHeight-textBoundingRect.height())/2
        elif self.vertical_align is Text.ALIGN_BOTTOM:		# BOTTOM ALIGN
            print "in bottom valign"
            yy = yy+ (selfHeight-textBoundingRect.height())
        else:
            print "in top valign"

        xx = xx*1.0/zoomFactor
        yy = yy*1.0/zoomFactor

        #print "returning %s,%s drawing offset" % (xx,yy)
        return 0, yy
        # return 0, 0
    """

    # END class Text
    pass


class Node(LocalGraphObject):

    ATTRIBUTES = LocalGraphObject.ATTRIBUTES + [
        'task',
        'statusUpdater',
        'shouldDisplayExecutionStatus',
        'shouldUpdateDisplay',
        'shouldUpdateImage',
        'shouldUpdateColors'
    ]

    PART_EXECUTE_STATUS = 'execute status'
    PART_DISPLAY_IMAGES = 'display images'

    PART_BORDER = "node border"
    PART_BACKGROUND = "node background"
    PART_LABEL = "node label"

    
    def __init__(self, nodeData, *args, **kwds):
        LocalGraphObject.__init__(self)

        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.inputPorts = {}
        self.outputPorts = {}
        self.parts = {
            "inputPorts":self.inputPorts,
            "outputPorts":self.outputPorts
            }

        
        self.nodeData = nodeData
        
        self.shouldUpdateDisplay(True)
        self.shouldUpdateImage(False)
        self.shouldUpdateColors(True)

        self.isClickable(True)
        self.isSelectable(True)
        return


    def _buildChildren(self):
        """
        returns the parts that are to be drawn
        """

        parts = []

        imageMap = self.parts.get(Node.PART_DISPLAY_IMAGES, {})
        if IMAGE_PARAMETER_SWEEP in imageMap:
            parts.append(imageMap[IMAGE_PARAMETER_SWEEP])
        if IMAGE_SELECTED_TRUE in imageMap:
            parts.append(imageMap[IMAGE_SELECTED_TRUE])
        if IMAGE_SELECTED_FALSE in imageMap:
            parts.append(imageMap[IMAGE_SELECTED_FALSE])

            
        parts.append(self.parts[Node.PART_BACKGROUND])
        parts.append(self.parts[Node.PART_LABEL])
        parts.extend(self.inputPorts.values())
        parts.extend(self.outputPorts.values())

        if self.shouldDrawExecuteStatus():
            parts.append(self.parts[Node.PART_EXECUTE_STATUS])

        return parts


    def partPrimitives(self):
        return self.children()

    def children(self):
        """
        A subclass can override this
        """
        if not hasattr(self, '_children'):
            # cache this so that we dont have to recompute 
            # for every redraw
            self._children = self._buildChildren()
            pass
        return self._children

    
    def configurePerformers(self, task):

        self.task(task)

        parentTask = task.parentTask()
        taskInfo = parentTask.getTaskInformation(task)
        values = taskInfo._values
        guiNode = self

        import mext.reaction.component as trellis
        
        class StatusUpdater(trellis.Component):
            @trellis.maintain
            def update(self):

                logging.debug("performer.update")
                
                if not hasattr(values, 'changed'):
                    return

                changes = getattr(values, 'changed')
                if not 'status' in changes:
                    return

                guiNode.updateExecuteStatus(changes['status'])
                task.automaton().contextManager().mainWindow().emit(
                    SIGNAL("OnTaskExecutionStatusChanged()"))

                        
                return
            # END class StatusUpdater
            pass

        statusUpdater = StatusUpdater()
        self.statusUpdater(statusUpdater)

        return

    def updateDisplay(self):
        if self.shouldUpdateColors():
            self.updateColors()

        if self.shouldUpdateImage():
            self.updateImage()
        return


    def shouldDrawExecuteStatus(self):
        dataNode = self.nodeData
        guiOptions = dataNode.guiOptions()
        return guiOptions.get('should draw execute status', True)

    def shouldDrawTemporalParameters(self):
        dataNode = self.nodeData
        guiOptions = dataNode.guiOptions()
        return guiOptions.get('should draw temporal parameters', True)
        

    def retrieveInputPortNamesFromDataObject(self):
        # getParametersByFilter

        filter = FilterModule.constructAndFilter()

        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter = FilterModule.EquivalenceFilter(ParameterModule.PORT_DIRECTION_INPUT),
                keyFunction=lambda x: x.portDirection()
            )
        )

        if not self.shouldDrawTemporalParameters():
            notTemporalFilter = FilterModule.constructNotFilter()
            notTemporalFilter.addFilter(
                FilterModule.ObjectKeyMatchesFilter(
                    filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_TEMPORAL),
                    keyFunction=lambda x: x.portType()
                )
            )
            filter.addFilter(notTemporalFilter)
            pass

        """
        notSideEffectFilter = FilterModule.constructNotFilter()
        notSideEffectFilter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.IdentityFilter(True),
                keyFunction=lambda x: x.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT)
            )
        )
        filter.addFilter(notSideEffectFilter)
        """

        parameters = self.nodeData.getParametersByFilter(filter)
        
        parameterNames = [x.id() for x in parameters]
        return parameterNames



    def retrieveOutputPortNamesFromDataObject(self):
        filter = FilterModule.constructOrFilter()

        outputValueFilter = FilterModule.ObjectKeyMatchesFilter(
            filter = FilterModule.EquivalenceFilter(ParameterModule.PORT_DIRECTION_OUTPUT),
            keyFunction=lambda x: x.portDirection()
        )

        if not self.shouldDrawTemporalParameters():
            f = FilterModule.constructAndFilter()
            f.addFilter(outputValueFilter)
            outputValueFilter = f
            
            notTemporalFilter = FilterModule.constructNotFilter()
            notTemporalFilter.addFilter(
                FilterModule.ObjectKeyMatchesFilter(
                    filter=FilterModule.EquivalenceFilter(ParameterModule.PORT_TYPE_TEMPORAL),
                    keyFunction=lambda x: x.portType()
                )
            )
            outputValueFilter.addFilter(notTemporalFilter)
            pass
        
        filter.addFilter(outputValueFilter)
        filter.addFilter(
            FilterModule.ObjectKeyMatchesFilter(
                filter=FilterModule.IdentityFilter(True),
                keyFunction=lambda x: x.getAttribute(ParameterModule.PORT_ATTRIBUTE_ISSIDEEFFECT)
            )
        )

        parameters = self.nodeData.getParametersByFilter(filter)
        parameterNames = [x.id() for x in parameters]
        return parameterNames

    

    def getResourceKeyForPort(self, portName):
        return 'gui port class'


    def setupPorts(self, nodePolicy):

        portPolicy = nodePolicy.portPolicy()

        inputPortNames = self.retrieveInputPortNamesFromDataObject()
        outputPortNames = self.retrieveOutputPortNamesFromDataObject()

        for portDirection, portNames, portMap in [
            (Port.PORT_DIRECTION_INPUT, inputPortNames, self.inputPorts),
            (Port.PORT_DIRECTION_OUTPUT, outputPortNames, self.outputPorts)]:

            positionCount = len(portNames)
            for index, portName in enumerate(portNames):

                resourceKey = self.getResourceKeyForPort(portName)
                portClass = self.contextManager().app().getResourceValue(
                    resourceKey, default=Port)
                port = portClass()
                port.name = portName
                port.node = self
                port.direction = portDirection
                port.positionIndex(index)
                port.positionCount(positionCount)

                port.initialSetup()

                portPolicy.setupDimensions(port)

                # we need to first know the port's dimensions
                # beefore we can set its position
                nodePolicy.setPortPosition(self, port)

                port.setupPrimitives(portPolicy)

                portMap[portName] = port
                pass
            pass

        return


    def setupPrimitives(self, nodePolicy):

        nodePolicy.setupDimensions(self)

        #font = nodePolicy.visualPolicy().font(
        #    PolicyModule.VisualPolicy.KEY_FONT_DEFAULT)
        #self.font = font

        background = self.createBackgroundPrimitive()

        nameLabel = self.createNamePrimitive(nodePolicy)

        self.parts[Node.PART_BACKGROUND] = background
        self.parts[Node.PART_LABEL] = nameLabel
        self.parts["colors"] = nodePolicy._colors

        self.setupPorts(nodePolicy)

        self.shouldUpdateDisplay(True)

        #font = nodePolicy.visualPolicy().font(
        #    PolicyModule.VisualPolicy.KEY_FONT_DEFAULT)

        # TODO: 
        # consolidate code that creates a new label
        executeLabel = Text()
        executeLabel.alpha = 1.0
        executeLabel.color = [1.0, 1.0, 1.0, 1.0]
        executeLabel.position = nodePolicy.getPositionForExecuteStatusLabel()
        executeLabel.size = [self.width, self.height, 0]
        # executeLabel.font = font

        # TODO:
        # this text should be dependent upon
        # whether there's a task associated with it
        executeLabel.text = "Uninitialized"
        self.parts[Node.PART_EXECUTE_STATUS] = executeLabel

        self.setupImageMap(nodePolicy)

        return

    def createNamePrimitive(self, nodePolicy):
        # Fit node to text size or simply to
        # what graphviz gives us, which may not look good
        # depending on the NodePolicy
        #stringBounds = font.getSize(node.nodeData.name)
        #nodeSize = (stringBounds[0] + 2*10, stringBounds[1] + 2*10)
        # nodeSize = self.getBounds()


        # initialize the name label
        nameLabel = Text()
        nameLabel.alpha = 1.0
        nameLabel.color = [1.0, 1.0, 1.0, 1.0]
        nameLabel.position = nodePolicy.getPositionForNameLabel()
        nameLabel.size = [self.width, self.height, 0]
        # nameLabel.font = self.font
        nameLabel.text = self.nodeData.name()
        return nameLabel


    def createBackgroundPrimitive(self, size=None):

        if size is None:
            size = self.getBounds()
        background = Rect(0, 0, size[0], size[1])

        background.size = size
        background.position = [0, 0, 0]


        background.corner_mode = True
        return background

    def getBounds(self):
        return (self.width, self.height)


    def getParameterSweepImage(self):
        return 'ps unselected.png'
    
    def getImageSelectedFile(self):
        return 'node selected.png'

    def getImageNotSelectedFile(self):
        return 'node unselected.png'

    def setupImageMap(self, nodePolicy):
        self._setupImageMap(nodePolicy)
        self.setImageSelected(False)
        return


    def setImageParameterSweep(self, isParameterSweep):
        imageMap = self.parts[Node.PART_DISPLAY_IMAGES]

        imageMap[IMAGE_PARAMETER_SWEEP].visible(isParameterSweep)
        return

    def setImageSelected(self, isSelected):
        imageMap = self.parts[Node.PART_DISPLAY_IMAGES]

        imageMap[IMAGE_SELECTED_TRUE].visible(isSelected)
        imageMap[IMAGE_SELECTED_FALSE].visible(not isSelected)
            
        return


    def _setupImageMap(self, nodePolicy):

        self.shouldUpdateImage(True)
        self.shouldUpdateColors(False)

        contextManager = nodePolicy.contextManager()
        app = contextManager.app()

        imageSelectedPath = os.path.join(contextManager.imagePath(),
                                         self.getImageSelectedFile())
        imageNotSelectedPath = os.path.join(contextManager.imagePath(),
                                            self.getImageNotSelectedFile())
        imageParameterSweepPath = os.path.join(contextManager.imagePath(),
                                           self.getParameterSweepImage())
        
        imageSelected = self.createBackgroundPrimitive()
        imageSelected._texture_image_file = imageSelectedPath
        imageSelected.use_texture = True

        imageNotSelected = self.createBackgroundPrimitive()
        imageNotSelected._texture_image_file = imageNotSelectedPath
        imageNotSelected.use_texture = True

        size = self.getBounds()
        diffX = size[0]/12.0
        diffY = -1*diffX
        imageParameterSweep = self.createBackgroundPrimitive(size=size)
        imageParameterSweep._texture_image_file = imageParameterSweepPath
        imageParameterSweep.use_texture = True

        # NOTE:
        # this is a hack to get the PS image drawing
        imageParameterSweep.offsetPosition = (diffX, diffY, 0)

        # now specify the images for selected/unselected
        imageMap = {
            IMAGE_SELECTED_TRUE:imageSelected,
            IMAGE_SELECTED_FALSE:imageNotSelected,
            IMAGE_PARAMETER_SWEEP:imageParameterSweep
        }

        self.parts[Node.PART_DISPLAY_IMAGES] = imageMap
        self.parts[Node.PART_BACKGROUND].visible(False)

        generator = TaskModule.getTaskGeneratorForDefinition(self.nodeData)
        isParameterSweep = isinstance(
            generator, TaskModule.ParameterSweepTaskGenerator)
        self.setImageParameterSweep(isParameterSweep)
        
        return


    def updateExecuteStatus(self, status):

        executeLabel = self.parts[Node.PART_EXECUTE_STATUS]
        executeLabel.text = status

        # reset the cached values
        # so that they will be recached
        if hasattr(self, '_children'):
            delattr(self, '_children')

        return


    def getPosition(self):
        return (self.x, self.y)

    def setPosition(self, x, y):
        self.x = x
        self.y = y
        return


    def updateColors(self):

        colors = self.parts["colors"]
        background = self.parts[Node.PART_BACKGROUND]
        text = self.parts[Node.PART_LABEL]
        if self.isSelected():
            logging.debug("updating for selected node")
            background.borderColor = colors[PolicyModule.NodePolicy.KEY_BORDER_SELECTION_TRUE]
            background.color = colors[PolicyModule.NodePolicy.KEY_BACKGROUND_SELECTION_TRUE]
            text.color = colors[PolicyModule.NodePolicy.KEY_TEXT_SELECTION_TRUE]
        else:
            logging.debug("updating for unselected node")
            background.borderColor = colors[PolicyModule.NodePolicy.KEY_BORDER_SELECTION_FALSE]
            background.color = colors[PolicyModule.NodePolicy.KEY_BACKGROUND_SELECTION_FALSE]
            text.color = colors[PolicyModule.NodePolicy.KEY_TEXT_SELECTION_FALSE]
            pass

        return


    def updateImage(self):
        self.setImageSelected(self.isSelected())
        return


    def intersectsRect(self, rect):

        if (rect.x+rect.width > self.x) and \
           (rect.x < self.x+self.width) and \
           (rect.y+rect.height > self.y) and \
           (rect.y < self.y+self.height):
            return True 

        return False


    def overlaps(self, point):
        if point.x < self.x:
            return False
        if point.x > self.x+self.width:
            return False
        if point.y < self.y:
            return False
        if point.y > self.y+self.height:
            return False
        return True


    # END class Node
    pass


class NestNode(Node):


    def getImageSelectedFile(self):
        return 'nest selected.png'

    def getImageNotSelectedFile(self):
        return 'nest unselected.png'

    # END class NestNode
    pass


class BranchNode(Node):

    # TODO:
    # implement the generation of the branch node image
    # and have this class return that value
    
    # END class BranchNode
    pass


class LoopNode(Node):

    # TODO:
    # implement the generation of the loop node image
    # and have this class return that value
    
    # END class LoopNode
    pass



class Edge(LocalGraphObject):

    DEFAULT_WIDTH = 3
    DEFAULT_WIDTH_SELECTED = 4

    ATTRIBUTES = LocalGraphObject.ATTRIBUTES + []

    def __init__(self, dataPath=None, guiPath=None):
        LocalGraphObject.__init__(self)

        self._dataPath = dataPath
        self.guiPath = guiPath

        self.isSelectable(True)
        self.isClickable(True)
        return


    def isDynamic(self, value=None):
        if value is not None:
            self._isDynamic = value
        if not hasattr(self, '_isDynamic'):
            self._isDynamic = False
        return self._isDynamic

    def getWidth(self):
        if self.isSelected():
            return Edge.DEFAULT_WIDTH_SELECTED
        return Edge.DEFAULT_WIDTH

    def getColor(self):
        if self.isDynamic():
            return self._dynamicColor
        if self.isSelected():
            return self._selectedColor
        return self._unselectedColor


    def inputPort(self):
        return self.guiPath[1]

    def outputPort(self):
        return self.guiPath[-2]

    def inputNode(self):
        return self.guiPath[0]

    def outputNode(self):
        return self.guiPath[-1]

    def sourceNode(self):
        return self.inputNode()

    def targetNode(self):
        return self.outputNode()


    def intersectsRect(self, rect):
        # contains = any([rect.containsPoint(Point(*x)) for x in self.points])
        # return contains
        # def intersectsLine(self, line):
        for startPoint, endPoint in zip(self.points[:-1], self.points[1:]):
            line = Line()
            line.points = [startPoint, endPoint]
            if rect.intersectsLine(line):
                return True
        if any([rect.containsPoint(x) for x in self.points]):
            return True
        return False


    def setupPrimitives(self, edgePolicy):
        self.children = []

        try:
            inputPort = self.inputPort()
            outputPort = self.outputPort()

            point1 = Point(
                self.inputNode().x + inputPort.x + inputPort.width/2, 
                self.inputNode().y + inputPort.y + inputPort.height/2)
            point2 = Point(
                self.outputNode().x + outputPort.x + outputPort.width/2, 
                self.outputNode().y + outputPort.y + outputPort.height/2)
        except AttributeError, e:
            logging.error(e)
            return

        line = Line()
        points = edgePolicy.createPath((point1.x, point1.y),
                                       (point2.x, point2.y))
        pointObjects = [Point(x, y, z) for x,y,z in points]
        line.points= pointObjects
        self.points = pointObjects
        self.children.append(line)
        edgePolicy.updateEdge(self)

        self.setupColors(edgePolicy)
        return



    def shouldUpdateDisplay(self, value):
        # no need for edges to update for now
        return

    def updateDisplay(self):
        # no need for edges to update for now
        return

    def setupColors(self, edgePolicy):

        visualPolicy = edgePolicy.visualPolicy()

        self._selectedColor = visualPolicy.color(
            PolicyModule.EdgePolicy.KEY_SELECTION_TRUE)
        self._unselectedColor = visualPolicy.color(
            PolicyModule.EdgePolicy.KEY_SELECTION_FALSE)
        self._dynamicColor = visualPolicy.color(
            PolicyModule.EdgePolicy.KEY_DYNAMIC)

        return

    def overlaps(self, point):
        rect = Rect(
            point.x-3, point.y-3,
            6, 6)
        return self.intersectsRect(rect)


    # END class Edge
    pass


class Port(LocalGraphObject):

    ATTRIBUTES = LocalGraphObject.ATTRIBUTES + [
        'positionIndex',
        'positionCount',
        ]

    PART_BACKGROUND = "port background"

    PORT_DIRECTION_INPUT = 1
    PORT_DIRECTION_OUTPUT = 2


    def __init__(self, *args, **kwds):

        LocalGraphObject.__init__(self)

        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.children = []
        self.edge = None
        self.basePosition = (0, 0)

        self.isClickable(True)
        self.isSelectable(False)
        return


    def createBackgroundPrimitive(self, portPolicy):
        
        dataNode = self.node.nodeData
        contextManager = portPolicy.contextManager()
        parameterId = self.name
        
        parameter = dataNode.getParameter(parameterId)

        background = None
        if self.dataType == ParameterModule.PORT_TYPE_TEMPORAL:
            background = Circle(0, 0, self.width/2.0)
            pass
        else:
            background = Rect(0, 0, self.width, self.height)
            pass
        
        background.position = [0, 0, 0]
        background.size = [self.width, self.height, 1.0]
        background.corner_mode = True
        
        background.color = None
        background.borderColor = [0.5, 0.5, 0.0, 1.0]
        
        return background
    
    
    def initialSetup(self):
        
        dataNode = self.node.nodeData
        parameterId = self.name
        parameter = dataNode.getParameter(parameterId)
        self.dataType = parameter.portType()
        
        return
    

    def overlaps(self, point):
        """
        TODO:
        This is copied from Node.Node
        Need to remove duplication
        """

        selfX = self.x + self.node.x
        selfY = self.y + self.node.y

        if point.x < selfX:
            return False
        if point.x > selfX+self.width:
            return False
        if point.y < selfY:
            return False
        if point.y > selfY+self.height:
            return False
        return True

    
    def setupPrimitives(self, portPolicy):
        
        background = self.createBackgroundPrimitive(portPolicy)
        
        self.parts = {}
        self.parts[Port.PART_BACKGROUND] = background

        self.children.append(background)

        return


    def updateColors(self):

        # No color policy for Ports set up yet 
        #colors = self.parts["colors"]

        background = self.parts[Port.PART_BACKGROUND]
        if self.edge is not None:
            background.borderColor = [0.8, 0.8, 0.0, 1.0]
            # background.color = [0.5, 0.5, 0.0, 1.0]
        else:
            background.borderColor = [0.8, 0.8, 0.0, 1.0]
            background.color = None
            pass

        return


    # END class Port
    pass



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/qt/edit definition dialog.ui'
#
# Created: Thu Mar  4 16:11:06 2010
#      by: PyQt4 UI code generator 4.7
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(409, 683)
        Dialog.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.buttonBox = QtGui.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(90, 640, 291, 32))
        self.buttonBox.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Apply|QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.checkBoxStageable = QtGui.QCheckBox(Dialog)
        self.checkBoxStageable.setGeometry(QtCore.QRect(130, 180, 121, 20))
        self.checkBoxStageable.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.checkBoxStageable.setObjectName("checkBoxStageable")
        self.label = QtGui.QLabel(Dialog)
        self.label.setGeometry(QtCore.QRect(21, 30, 71, 16))
        self.label.setObjectName("label")
        self.label_2 = QtGui.QLabel(Dialog)
        self.label_2.setGeometry(QtCore.QRect(21, 70, 71, 16))
        self.label_2.setObjectName("label_2")
        self.label_3 = QtGui.QLabel(Dialog)
        self.label_3.setGeometry(QtCore.QRect(20, 110, 101, 16))
        self.label_3.setObjectName("label_3")
        self.labelExecutablePath = QtGui.QLabel(Dialog)
        self.labelExecutablePath.setGeometry(QtCore.QRect(20, 150, 101, 16))
        self.labelExecutablePath.setObjectName("labelExecutablePath")
        self.labelJarFile = QtGui.QLabel(Dialog)
        self.labelJarFile.setGeometry(QtCore.QRect(20, 220, 62, 16))
        self.labelJarFile.setObjectName("labelJarFile")
        self.labelJarClass = QtGui.QLabel(Dialog)
        self.labelJarClass.setGeometry(QtCore.QRect(20, 260, 62, 16))
        self.labelJarClass.setObjectName("labelJarClass")
        self.labelStaticArgs = QtGui.QLabel(Dialog)
        self.labelStaticArgs.setGeometry(QtCore.QRect(20, 300, 71, 16))
        self.labelStaticArgs.setObjectName("labelStaticArgs")
        self.label_8 = QtGui.QLabel(Dialog)
        self.label_8.setGeometry(QtCore.QRect(20, 350, 81, 16))
        self.label_8.setObjectName("label_8")
        self.tableParameters = QtGui.QTableWidget(Dialog)
        self.tableParameters.setGeometry(QtCore.QRect(20, 380, 361, 191))
        self.tableParameters.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.tableParameters.setObjectName("tableParameters")
        self.tableParameters.setColumnCount(2)
        self.tableParameters.setRowCount(0)
        item = QtGui.QTableWidgetItem()
        self.tableParameters.setHorizontalHeaderItem(0, item)
        item = QtGui.QTableWidgetItem()
        self.tableParameters.setHorizontalHeaderItem(1, item)
        self.buttonRemove = QtGui.QPushButton(Dialog)
        self.buttonRemove.setGeometry(QtCore.QRect(290, 580, 91, 32))
        self.buttonRemove.setObjectName("buttonRemove")
        self.buttonEdit = QtGui.QPushButton(Dialog)
        self.buttonEdit.setGeometry(QtCore.QRect(200, 580, 91, 32))
        self.buttonEdit.setObjectName("buttonEdit")
        self.buttonAdd = QtGui.QPushButton(Dialog)
        self.buttonAdd.setGeometry(QtCore.QRect(110, 580, 91, 32))
        self.buttonAdd.setObjectName("buttonAdd")
        self.line = QtGui.QFrame(Dialog)
        self.line.setGeometry(QtCore.QRect(20, 330, 361, 16))
        self.line.setFrameShape(QtGui.QFrame.HLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.line.setObjectName("line")
        self.line_2 = QtGui.QFrame(Dialog)
        self.line_2.setGeometry(QtCore.QRect(20, 620, 361, 16))
        self.line_2.setFrameShape(QtGui.QFrame.HLine)
        self.line_2.setFrameShadow(QtGui.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.comboBoxExecutableType = QtGui.QComboBox(Dialog)
        self.comboBoxExecutableType.setGeometry(QtCore.QRect(130, 100, 251, 31))
        self.comboBoxExecutableType.setObjectName("comboBoxExecutableType")
        self.textEditName = QtGui.QLineEdit(Dialog)
        self.textEditName.setGeometry(QtCore.QRect(130, 21, 251, 31))
        self.textEditName.setObjectName("textEditName")
        self.textEditDescription = QtGui.QLineEdit(Dialog)
        self.textEditDescription.setGeometry(QtCore.QRect(130, 60, 251, 31))
        self.textEditDescription.setObjectName("textEditDescription")
        self.textEditExecutablePath = QtGui.QLineEdit(Dialog)
        self.textEditExecutablePath.setGeometry(QtCore.QRect(130, 140, 251, 31))
        self.textEditExecutablePath.setObjectName("textEditExecutablePath")
        self.textEditJarFile = QtGui.QLineEdit(Dialog)
        self.textEditJarFile.setGeometry(QtCore.QRect(130, 210, 251, 31))
        self.textEditJarFile.setObjectName("textEditJarFile")
        self.textEditJarClass = QtGui.QLineEdit(Dialog)
        self.textEditJarClass.setGeometry(QtCore.QRect(130, 250, 251, 31))
        self.textEditJarClass.setObjectName("textEditJarClass")
        self.textEditStaticArgs = QtGui.QLineEdit(Dialog)
        self.textEditStaticArgs.setGeometry(QtCore.QRect(130, 290, 251, 31))
        self.textEditStaticArgs.setObjectName("textEditStaticArgs")

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtGui.QApplication.translate("Dialog", "Edit definition", None, QtGui.QApplication.UnicodeUTF8))
        self.checkBoxStageable.setText(QtGui.QApplication.translate("Dialog", "Stageable", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("Dialog", "Name", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("Dialog", "Description", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("Dialog", "Executable type", None, QtGui.QApplication.UnicodeUTF8))
        self.labelExecutablePath.setText(QtGui.QApplication.translate("Dialog", "Executable path", None, QtGui.QApplication.UnicodeUTF8))
        self.labelJarFile.setText(QtGui.QApplication.translate("Dialog", "Jar file", None, QtGui.QApplication.UnicodeUTF8))
        self.labelJarClass.setText(QtGui.QApplication.translate("Dialog", "Jar class", None, QtGui.QApplication.UnicodeUTF8))
        self.labelStaticArgs.setText(QtGui.QApplication.translate("Dialog", "Static args", None, QtGui.QApplication.UnicodeUTF8))
        self.label_8.setText(QtGui.QApplication.translate("Dialog", "Parameters", None, QtGui.QApplication.UnicodeUTF8))
        self.tableParameters.horizontalHeaderItem(0).setText(QtGui.QApplication.translate("Dialog", "Name", None, QtGui.QApplication.UnicodeUTF8))
        self.tableParameters.horizontalHeaderItem(1).setText(QtGui.QApplication.translate("Dialog", "Direction", None, QtGui.QApplication.UnicodeUTF8))
        self.buttonRemove.setText(QtGui.QApplication.translate("Dialog", "Remove", None, QtGui.QApplication.UnicodeUTF8))
        self.buttonEdit.setText(QtGui.QApplication.translate("Dialog", "Edit", None, QtGui.QApplication.UnicodeUTF8))
        self.buttonAdd.setText(QtGui.QApplication.translate("Dialog", "Add", None, QtGui.QApplication.UnicodeUTF8))


if __name__ == "__main__":
    import sys
    app = QtGui.QApplication(sys.argv)
    Dialog = QtGui.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())

